package fi.jyu.mit.Music;

import java.util.LinkedList;
import java.util.List;


/**
 * Puskuroitu äänijärjestelmä, jonka tarkoitus on tukea mahdollisuutta 
 * syättää äänijärjestelmälle nuotteja, jotka voidaan toistaa
 * joko pääohjelman säikeessä tai uudessa säikeessä.
 * 
 * Tälläin käskyt play ja playNote eivät toista saamiansa nuotteja heti,
 * vaan ne puskuroidaan, eli varastoidaan. Soittaminen aloitetaan metodeilla
 * run tai runAsync.
 * 
 * Metodilla run äänijärjestelmä aloittaa soittamisen nykyisessä säikeessä,
 * kun taas runAsync tekee uuden säikeen joka sitten pyärii niin kauan kunnes
 * soitto on loppunut.
 * 
 * Asynkronisen soittamisen voi aina lopettaa käskyllä stop.
 */
public class BufferedMidiPlayer extends MidiPlayer {
    /**
     * Peritty luokka säiettä varten.
     */
    protected class SoundSystemThread extends Thread {
        @Override
        public void run() {
            boolean halted = false;
            // Luodaan säikeelle kopio nuottipuskurista 
            List<Note> localBuffer = new LinkedList<Note>(getNoteBuffer());
            while (!localBuffer.isEmpty()) {
                if (isHalt()) {
                    halted = true;
                    break;
                }
                Note n = localBuffer.remove(0);
                try  {
                    if (n.isDelay()) {
                        Thread.sleep(n.getLength());
                        message(currentChannel + " Tauko         " + n.message());
                    } else {
                        channels[currentChannel].noteOn(n.getNote(), n.getVelocity());
                        message( currentChannel + " Soitin nuotin " + n.message());
                        Thread.sleep(n.getLength());
                        channels[currentChannel].noteOff(n.getNote());
                    }
                }
                catch (InterruptedException e) {
                    if ( !n.isDelay() ) channels[currentChannel].noteOff(n.getNote());
                }
            }
            if (halted) {
                message("Soitto keskeytettiin!");
            } else {
                message("Soitto päättynyt!");
            }
        }
    }
    
    /**
     * Pysäytyslippu.
     */
    private boolean halt = true;
    
    /**
     * @return onko pysätetty
     */
    protected boolean isHalt() {
        return halt;
    }
    
    /**
     * Nuottipuskuri. Soitetaan tulojärjestyksessä.
     */
    private List<Note> noteBuffer;
    
    
    /**
     * @return nuottipuskuri
     */
    protected List<Note> getNoteBuffer() {
        return noteBuffer;
    }
    
    /**
     * alustetaan tyhjällä nuottilistalla
     */
    public BufferedMidiPlayer() {
        noteBuffer = new LinkedList<Note>();
    }
    
    
    /**
     * Alustetaan tietylle kanavalle
     * @param channel kanava jolle alustetaan
     */
    public BufferedMidiPlayer(int channel) {
        this();
        selectChannel(channel);
    }
    
    
    /**
     * Luo kopion puskuroiduista nuoteista
     * @return kopio nuoteista
     */
    public List<Note>getCopyOfNotes() {
        List<Note> notes = new LinkedList<Note>();
        for (Note n:noteBuffer)
            notes.add(n);
        return notes;
    }
    
    /**
     * Palauttaa nuotit viitteenä jolloin niiden muuttaminen muuttaa soitettavia nuotteja
     * @return nuotit
     */
    public List<Note>getNotes() {
        return noteBuffer;
    }
    
    /**
     * Tallentaa nuotin puskuriin.
     */
    @Override
    public void midiSoundNote(int note, int length, int velocity) {
        noteBuffer.add(new Note(note, length, velocity));
    }
    
    @Override
    protected void rest(final int barLength) {
        // Ei negatiivisia taukoja
        if (barLength < 0)
            return;
        noteBuffer.add(new Note.Delay(barLength));
    }
    
    
    /**
     * Aloittaa soiton kutsujan säikeessä, toisin sanoen keskeyttää pääohjelman toiminnan.
     */
    public void run() {
        while (!noteBuffer.isEmpty()) {
            Note n = noteBuffer.remove(0);
            try  {
                if ( n.isDelay() ) {
                    Thread.sleep(n.getLength());
                    message("Tauko         " + n.message());
                } else {
                    channels[currentChannel].noteOn(n.getNote(),n.getVelocity());
                    message("Soitin nuotin " + n.message());
                    Thread.sleep(n.getLength());
                    channels[currentChannel].noteOff(n.getNote());
                }
            }
            catch (InterruptedException e) {
                if ( !n.isDelay() ) channels[currentChannel].noteOff(n.getNote());
            }
        }
    }
    
    /**
     * Aloittaa asynkronisen soiton, eli soitto ei keskeytä pääohjelman toimintaa.
     */
    public void runAsync() {
        SoundSystemThread sst = new SoundSystemThread();
        halt = false;
        sst.start();
    }
    
    // git+ssh://git@github.com/ane/SoundSystem.git
    
    /**
     * Pysäyttää asynkronisen soiton.
     */
    public void stop() {
        halt = true;
    }
}
