package fi.jyu.mit.Music;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import javax.sound.midi.*;

/**
 * MidiPlayer-luokka sisältää varsinaisen äänijärjestelmän yksinkertaisen MIDI-datan toistoa varten.
 * äänijärjestelmä kykenee soittamaan yhdellä instrumentilla nuoteista koostuvaa soittoa. Kaikki GM-1 
 * standardin tukemat instrumentit ovat käytässä. Standardin määritykset läytyvät <a href="http://www.midi.org/techspecs/gm1sound.php">spesifikaatiosta</a>.
 *
 * Oktaavi, ks: esim: http://fi.wikipedia.org/wiki/Oktaavi
 * 
 * @see BufferedMidiPlayer
 */
public class MidiPlayer implements BasicMidiPlayer {  
    /**
     * Nuotin kestot tahdin desimaaliosissa. 
     */
    @SuppressWarnings("javadoc")
    public enum NoteDuration {
        DOUBLE(2.0),
        EIGHTH(0.125),
        HALF(0.5),
        QUARTER(0.25),
        SIXTEENTH(0.0625),
        SIXTYFOURTH(0.015625),
        THIRTYSECOND(0.03125),
        WHOLE(1.0);
        
        private final double length;
        
        NoteDuration(double length) {
            this.length = length;
        }
        
        public double length() { return this.length; }
    }
    
    /** Oletusnopeus */
    public static final int DEFVELOCITY = 64;
    /** Oletusoktaavia */
    public static final int DEFOCTAVE = 4;
    /** Taukoa vastaava merkkijono */
    public static final String DELAYST = "-";
        
    
    /**
     * Lista nuoteista.
     */
    public static final String[] ALLNOTES = {
        "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "Bb", "B"
    };
    /** Perusnuotit */
    public static final String basicNotes  = "C D EF G A B";
    /** Perusnuotit */
    public static final String basicNotes2 = "C D EF G A H";
    
    /**
     * Tempo. Määräytyy BPM:n (beats per minute) mukaan, oletus on 120 eli 
     * 120 iskua minuutissa, tällöin yhden iskun pituus on 500 ms.
     */
    private int beatLength = 500; // 120 BPM

    /**
     * Nopeus jolla "kosketinta" lyödään, 64 on oletus
     */
    private int velocity = DEFVELOCITY; 
    
    /**
     * Tulostetaanko debuggausviestejä. Oletuksena on, että ei, mutta toggleVerbosella sen voi muuttaa.
     */
    private boolean verbose = false;
    
    /**
     * Äänipankkitiedosto latausta varten.
     */
    private final String remoteSoundbankUrl = "http://java.sun.com/products/java-media/sound/soundbank-min.gm.zip";
    
    /**
     * Taulukko midikanavia varten.
     */
    protected MidiChannel[] channels;
    
    /**
     * Nykyinen käytässä oleva kanava. Oletuksena 0.
     */
    protected int currentChannel = 0;
    
    /**
     * Taulukko instrumentteja varten.
     */
    protected Instrument[] instruments;

    
    /**
     * Nykyinen instrumentti;
     */
    private int instrument = 1;
    

    /**
     * Iskuja minuutissa
     */
    private int bpm = 120;
    
    
    /**
     * Oletusoktaavi, josta nuotit soitetaan
     */
    private int defOctave = DEFOCTAVE;
    
    /**
     * @return oletusoktaavi
     */
    public int getDefOctave() {
        return defOctave;
    }

    /**
     * @param defocatve asetettava oletusoktaavi
     */
    public void setDefOctave(int defocatve) {
        this.defOctave = defocatve;
    }

    /**
     * Syntetisaattori.
     */
    protected Synthesizer synth;
    
    /**
     * Muodostaa äänijärjestelmän. Alustaa syntetisaattorin ym.
     */
    public MidiPlayer() {
        this.init();
    }
    
    /**
     * Muodostaa äänijärjestelmän, kuitenkaan syntetisaattoria alustamatta. Tämä tarpeen jos ei
     * halua viedä syntesaattoria järjestelmällä joissa niitä on vain yksi. 
     * @param init Alustustehto. Jos true, niin syntetisaattori käynnistetään, jos false, niin ei.
     */
    public MidiPlayer(final boolean init) {
        if (init)
            this.init();
    }
       
    /**
     * Jäsentää merkkijonon nuotit & tauot yksittäisiksi nuotteiksi.
     * Jonossa saa olla ylmääräisiä välilyäntejä tai | merkkejä selkeyttämässä jakoa
     * ilman että ne sotkevat mitään. Esimerkkjä käytöstä ks alla olevat testit.
     * 
     * @param sequence Merkkijono nuoteista, esim. C#5 
     * @param defaultOctave Oletusoktaavi, jossa nuotit soitetaan, jos niille ei annettu oktaavia
     * @param length Yksittäisen nuotin pituus jos nuotille ei erikeen ole annettu pituutta
     * @param velocity1 voimakkuus, 64 on perus
     * @return jäsennetty nuottilista
     * @example
     * <pre name="test">
     * #import java.util.*;
     *   MidiPlayer mp = new MidiPlayer(); List<Note> notes;
     *   
     *   notes = mp.getNotes("-b -/4 -#*",0.25,4,64); // - = tauko, ylennys tai alennus ei haittaa 
     *   notes.get(0).isDelay()   === true;
     *   notes.get(1).isDelay()   === true;
     *   notes.get(2).isDelay()   === true;
     *   notes.get(0).getLength() === 500;
     *   notes.get(1).getLength() === 125;
     *   notes.get(2).getLength() === 1000;
     *   
     *   notes = mp.getNotes("C/E/4D*F*4A.F*1.2",0.25,4,64); //  /nro jakaa pituuden, *nro kertoo
     *   notes.get(0).isDelay()   === false;      
     *   notes.get(0).getLength() === 250;                   // / yksin on sama kuin /2
     *   notes.get(1).getLength() === 125;                   // /4 jakaa pituuden 4:llä
     *   notes.get(2).getLength() === 1000;                  // *  yksin on sama kuin *2
     *   notes.get(3).getLength() === 2000;                  // *4 kertoo pituuden 4:llä
     *   notes.get(4).getLength() ~~~ 500*1.5;               // kerrotaan pituus 1.5:lla
     *   notes.get(5).getLength() ~~~ 500*1.2;               // ja 1.2lla
     *
     *   notes = mp.getNotes("C C# D D# E F F# G G# A Bb B",0.25,4,64); // oktaavin kaikki nuotit
     *   for (int i=0; i<12; i++) {
     *     notes.get(i).getNote() === 60 + i;
     *   }
     *      
     *   notes = mp.getNotes("c c# d d# e f f# g g# a Bb B",0.25,4,64); // samat pienillä (paitsi B)
     *   for (int i=0; i<12; i++) {
     *     notes.get(i).getNote() === 60 + i;
     *   }
     *      
     *   notes = mp.getNotes("C Db D Eb E F Gb | G Ab A A# B",0.25,4,64); // ylennetään "eritavalla"
     *   for (int i=0; i<12; i++) {
     *     notes.get(i).getNote() === 60 + i;
     *   }   
     *   
     *   notes = mp.getNotes("HBhbBb",0.25,4,64);                         // H ja B ovat synonyymejä
     *   notes.get(0).getNote() === notes.get(1).getNote()
     *   notes.get(2).getNote() === notes.get(3).getNote()
     *     
     *   notes = mp.getNotes("C1C2C3C4C5C6C7C8",0.25,4,64); // kokeillaan C:tä eri oktaaveista
     *   notes.get(0).getNote() === 60 - 3*12  
     *   notes.get(1).getNote() === 60 - 2*12  
     *   notes.get(2).getNote() === 60 - 1*12  
     *   notes.get(3).getNote() === 60 - 0*12  
     *   notes.get(4).getNote() === 60 + 1*12  
     *   notes.get(5).getNote() === 60 + 2*12  
     *   notes.get(6).getNote() === 60 + 3*12  
     *   notes.get(7).getNote() === 60 + 4*12
     *     
     *   notes = mp.getNotes("E# F H# C5",0.25,4,64);          
     *   notes.get(0).getNote() === notes.get(1).getNote()  
     *   notes.get(2).getNote() === notes.get(3).getNote()
     *     
     *   notes = mp.getNotes("A#3* Hb3*",0.25,4,64);      // Ylennetty A on sama kuin alennettu H molemmat 2*
     *   notes.get(0).getNote() === notes.get(1).getNote()  
     *   notes.get(0).getLength() === notes.get(1).getLength()  
     * </pre>
     */
    public final List<Note> getNotes(final String sequence, final double length, final int defaultOctave, final int velocity1) {
        List<Note> notes = new ArrayList<Note>();
        final Pattern p = Pattern.compile("(([-acdefghACDEFGBH])([#b]?)([0-8]?)([/*.][0-8.]*)?)");
        final Matcher m = p.matcher(sequence);
        
        Note n;
        while (m.find()) {
            // Nuotti on toinen täsmäysjoukko (ensimmäinen on tauko)
            String note = m.group(2);
            // Ylennys/madallusmerkki
            String pitchMove = (m.group(3) != null) ? m.group(3) : "";
            // Oktaavi
            int octave = (!m.group(4).isEmpty()) ? Integer.parseInt(m.group(4)) : defaultOctave;
            double factor = 1.0;
            String sf = m.group(5);
            char cf = 0;
            if (sf != null && sf.length() > 0) cf = sf.charAt(0);
            if (cf == '/' && sf != null ) {
                factor = 0.5;
                String smult = sf.substring(1);
                if (smult.length() > 0) factor = 1.0 / Double.parseDouble(smult);
            }
            if (cf == '*' && sf != null) {
                factor = 2.0;
                String smult = sf.substring(1);
                if (smult.length() > 0) factor = Double.parseDouble(smult);
            }
            if (cf == '.')
                factor = 1.5;

            n = makeNote(note + pitchMove, length * factor, octave, velocity1);
            notes.add(n);
        }
        return notes;
    }
    
    /** 
     * Alustaa MIDI-syntetisaattorin.
     */
    public void init() {
        Soundbank s = null;
        try {
            synth = MidiSystem.getSynthesizer();
            synth.open();
            if ( verbose ) System.out.println("Ladattiin syntetisaattori " + synth);
            s = synth.getDefaultSoundbank();
            // Syntetisaattori ei saanut äänipankkia, joten ladataan sen etäpalvelimelta.
            if (s == null) {
                synth.close();
                downloadSoundbank(remoteSoundbankUrl);
                synth.open();
                s = synth.getDefaultSoundbank();
                // Jos ei vieläkään
                if (s == null) {
                    error("Lataaminen epäonnistui tai äänipankkitiedosto on epäkelpo. Ks. https://trac.cc.jyu.fi/projects/ohj1/wiki/music");
                }
            }
        } catch (MidiUnavailableException e) {
            e.printStackTrace();
            return;
        }
        if ( s!= null ) instruments = s.getInstruments();
        if (instruments == null) {
            error("Instrumenttien lataamisessa tapahtui virhe.");
        }
       
        synth.loadInstrument(instruments[1]); // piano
        channels = synth.getChannels();
        selectChannel(1);
    }
    
    /**
     * Lataa äänipankin Sunin sivuilta.
     * @param remote URL, josta ladataan
     */
    @SuppressWarnings("resource") // TODO: poikkeustapauksessa ei suljeta
    private void downloadSoundbank(String remote) {
        HttpURLConnection source = null;
        BufferedOutputStream bos = null;
        BufferedInputStream bis = null;
        final String zipFileName = "soundbank-min.gm.zip";
        try {
            File z = new File(zipFileName);
            // Jos zip-pakettia ei lÃ¶ydy jo valmiiksi, ladataan se
            if (!z.exists()) {
                URL url = new URL(remote);
                source = (HttpURLConnection) url.openConnection();
                source.connect();
                System.out.print("Haetaan äänipankki osoitteesta " + remote + " ... ");
                // Tarkista palvelimen vastaus
                String response = "HTTP " + source.getResponseCode() + " " + source.getResponseMessage();
                // Jos 404 (ei lÃ¶ydy)
                if (source.getResponseCode() != HttpURLConnection.HTTP_OK) {
                    System.err.println(response);
                    error("Äänipankkitiedoston latauksessa tapahtui virhe. Ks. https://trac.cc.jyu.fi/projects/ohj1/wiki/music");
                } 
                System.out.println(response);
                bis = new BufferedInputStream(source.getInputStream());
                bos = new BufferedOutputStream(new FileOutputStream(zipFileName));
                int count = 0;
                byte[] buf = new byte[8192]; // HTTP-puskurin suositeltu koko
                // Lataa tiedosto aina puskurin koko kerrallaan
                int fileSize = source.getContentLength(); // tiedoston koko palvelimella
                System.out.print("Ladataan " + zipFileName + " (" + fileSize/1000 + " kB) ... ");
                for (int bytesRead = bis.read(buf); bytesRead > 0; bytesRead = bis.read(buf)) {
                    count += bytesRead;
                    bos.write(buf, 0, bytesRead);
                }
                System.out.println(count/1000 + " kB ladattu");
                bis.close();
                bos.close();
                source.disconnect();
            }
            // Puretaan tiedosto
            ZipFile pkg = null;
            try {
                pkg = new ZipFile(new File(zipFileName));
                // Hae paketin sisältÃ¶
                Enumeration<? extends ZipEntry> entries = pkg.entries();
                while (entries.hasMoreElements()) {
                    ZipEntry ze = entries.nextElement();
                    // Pura vain .gm-päätteiset tiedostot
                    if (!ze.isDirectory() && ze.getName().endsWith(".gm")) {
                        System.out.println("Puretaan äänipankki " + ze.getName());
                        // Kopioi tiedoston sisältÃ¶
                        BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(ze.getName()));
                        BufferedInputStream in = new BufferedInputStream(pkg.getInputStream(ze));
                        int data;
                        while ((data = in.read()) != -1) {
                            out.write((byte) data);
                        }
                        out.close();
                        in.close();
                    }
                }
            } catch (ZipException ze) {
                System.err.println("Virhe avattaessa zip-tiedostoa: " + ze.getMessage());
            }
            finally {
                if ( pkg != null) pkg.close();
            }
        }
        catch (MalformedURLException mue) {
            error("Virheellinen URL");
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
    
    /**
     * Muuntaa nuotin, oktaavin ja voimakkuuden Note-luokan alkioksi. <b>Huom.! Nuotin perään ei tule oktaavia.</b>
     * Tauko on - -merkki.
     * @param note Nuotti tekstimuodossa, esim. C# tai Bb 
     * @param length Nuotin pituus tahdin kokonaisosina
     * @param octave Oktaavi (0-8)
     * @param velocity1 voimakkuus, 64 on normaali
     * @return Muunnettu nuotti.
     * @see Note
     * @example
     * <pre name="test">
     *   MidiPlayer mp = new MidiPlayer();
     *   Note note;
     *   
     *   note = mp.makeNote("C",0.5,4,64);
     *   note.getLength() === 1000;
     *   note.getNote() === 60;
     *   
     *   note = mp.makeNote("C#",0.5,4,64); // korotus
     *   note.getNote() === 61;
     *   
     *   note = mp.makeNote("H#",0.5,4,64); // korotettu H on seuraava C
     *   note.getNote() === 72;
     *   
     *   note = mp.makeNote("Bb",0.5,4,64); // Alennettu B (H)
     *   note.getNote() === 70;
     *   
     *   note = mp.makeNote("-",0.5,4,64); // Tauko
     *   note.isDelay() === true;
     *   
     *   note = mp.makeNote("-#",0.5,4,64); // Tauko, ylennys ei haittaa
     *   note.isDelay() === true;
     *   
     *   note = mp.makeNote("-b",0.5,4,64); // Tauko, alennus ei haittaa
     *   note.isDelay() === true;
     * </pre>
     */
    public final Note makeNote(final String note, final double length, final int octave, final int velocity1) {
        final int noteLength = (int)((this.beatLength * 4) * length);
        
        if ( octave < 0  || note.length() < 1 || DELAYST.indexOf(note.charAt(0)) >= 0 ) 
            return new Note.Delay(noteLength);
        
        char noteChar = Character.toUpperCase(note.charAt(0));
        char pitchMove = 0;
        if ( note.length() > 1 ) pitchMove = note.charAt(1);
        
        // Tarkistetaan että nuotit & oktaavit ovat oikein
        
        int noteIndex = basicNotes.indexOf(noteChar);
        if ( noteIndex < 0 ) noteIndex = basicNotes2.indexOf(noteChar);
        
        if ( noteIndex < 0 ) {
            error("Virheellinen nuotti: " + note);
            return null;
        } 
        if (octave < 0 || octave > 8) {
            error("Virheellinen oktaavi: " + octave);
            return null;
        }
        if (length < 0.0625) {
            error("Virheellinen nuotin kesto, " + length + " minimi: 0.0625 - 1");
            return null;
        }
        int base = 24;
        int resultNote = 0;
        
        if ( pitchMove == '#' ) noteIndex++;
        if ( pitchMove == 'b' ) noteIndex--;
        
        // Valitaan oktaavi. Ensimmäisesä 0-oktaavissa on vain kolme nuottia (A, Bb, B), joten
        // indeksin kaksi yli menevät (C-G#) tuottavat virheen.
        if (octave == 0) {
            if (noteIndex < 9) {
                error("Virhe: nollaoktaavissa soivat vain A, Bb ja B, koitit: " + note);
            } else {
                resultNote = 21 + (noteIndex - 9);
            }
        } else {
            resultNote = base + ((octave - 1) * 12) + noteIndex;
        }
        return new Note(resultNote, noteLength, velocity1);
    }
    

    /**
     * Muuntaa nuotin, oktaavin ja voimakkuuden Note-luokan alkioksi. <b>Huom.! Nuotin perään ei tule oktaavia.</b>
     * @param note Nuotti tekstimuodossa, esim. "C#"
     * @param length Nuotin pituus tahdin kokonaisosina
     * @param octave Oktaavi (0-8)
     * @return muutettu nuotti
     */
    public final Note makeNote(final String note, final double length, final int octave) {
       return makeNote(note,length,octave,this.velocity); 
    }
    
    /**
     * Soittaa kanavalta MIDI-nuotin (21-108). 
     * @param note Nuotti (21-108)
     * @param length Kesto millisekunteina
     * @param velocity1 Nuotin nk. painovoimakkuus. Oletus on 64.
     */
    public void midiSoundNote(final int note, final int length, final int velocity1)
    {
        checkSynth("midiSoundNote");
        try 
        {
            channelOn(note, velocity1);
            message("Soitin nuotin " + Note.message(note, length, velocity1));
            Thread.sleep(length);
            channelOff(note);
        }
        catch (InterruptedException e)
        {
            channelOff(note);
        }
    }
    
    /**
     * Aloittaa kanavalta MIDI-nuotin (21-108). 
     * @param note Nuotti (21-108)
     * @param length Kesto millisekunteina
     * @param velocity1 Nuotin nk. painovoimakkuus. Oletus on 64.
     */
    public void midiStartNote(final int note, final int length, final int velocity1)
    {
        checkSynth("midiStartNote");
        channelOn(note, velocity1);
        message("Aloitin nuotin " + Note.message(note, length, velocity1));
    }
    
    /**
     * Aloittaa kanavalta MIDI-nuotin (21-108). 
     * @param note Nuotti (21-108)
     * @param length Kesto millisekunteina
     * @param velocity1 Nuotin nk. painovoimakkuus. Oletus on 64.
     */
    public void midiEndNote(final int note, final int length, final int velocity1)
    {
        checkSynth("midiEndNote");
        message("Lopetin nuotin " + Note.message(note, length, velocity1));
        channelOff(note);
    }
    
    /**
     * Toistaa nuotteja. Nuottien kesto oletetaan neljäsosanuoteiksi.
     * @param sequence Nuotit
     */
    @Override
    public void play(final String sequence){
        play(sequence, 0.25, getDefOctave());
    }
    
    /**
     * Toistaa nuotteja tietyssä oktaavissa neljäsosanuotteina.
     * @param sequence Nuotit
     */
    @Override
    public void play(final String sequence, final double length) {
        play(sequence, length, getDefOctave());
    }
    
    /**
     * Toistaa merkkijonon neljäsosanuotteja perusoktaavissa (4), esim. "C5CC5CBbAA#". Huomaa, että "Bb" on aina annettava
     * sellaisenaan, sillä parseri ei osaa erottaa onko kyseessä "bb" vai "BB", parseri olettaa että tuolloin on kaksi
     * B-nuottia.
     * 
     * 
     * Syntaksi yksittäiselle nuotille on seuraava: ABCD, jossa
     * <ul>
     *     <li><i>A</i> on itse nuotti <b>(pakollinen)</b></li>
     *     <li><i>B</i> on nuotille ylennysmerkki # tai alennusmerkki b</li>
     *     <li><i>C</i> on nuotille oktaavi</li>
     *     <li><i>D</i> on nuotin kesto suhteessa peruspituuteen joko /jakaja tai *kerroin. Jos jakaja tai kerroin puuttu, se on 2</li>
     * </ul>
     * Mikä tahansa osista B, C tai D voi puuttua.
     * <pre>
     *  Esim:
     *    C#A   on ylennetty C ja A, oktaavista 4, kesto perusnuotin verran (yleensä siis 1/4 nuotti)
     *    C5    C 5:stä oktaavista
     *    C*2   on perusnuottia 2 pidempi nuotti (yleensä siis puolinuotti)
     *    C/4   C jonka kesto on 1/4 perusnuotista (siis 1/16 nuotti yleensä)
     *    A/    A jonka kesto on puoleta perusnuotista (yleensä siis 1/8 nuotti)
     *    C#5*4 4 kertaa normaali pidempi ylennetty C 5. oktaavista    
     *    D.    1.5 kertaa normaalia pidempi D, eli sama kuin D*1.5
     * </pre> 
     * @param sequence Nuotit merkkijonona, jonossa olevat välilyönnit ja | eivät haittaa
     *                 kunhan niitä ei laiteta nuotin sisälle
     * @param octave Oletusoktaavi.  Jos nuotille ei ole omaa oktaavia, käytettään tätä
     * @param length Yksittäisen nuotin kesto, tähän suhteutetaan muut jonon nuotit
     * @param velocity1 voimakkuus, 64 on normaali
     */
    @Override
    public void play(final String sequence, final double length, final int octave, final int velocity1) {
        play(getNotes(sequence, length, octave, velocity1));
    }

    
    @Override
    public void play(final String sequence, final double length, final int octave) {
        play(sequence, length, octave, this.velocity);
    }
    

    /**
     * Soittaa nuotit nuottien listasta
     * @param notes soitettavat nuotit
     */
    @Override
    public void play(final List<Note> notes) {
        for (Note n : notes) {
            // Tauko
            if (n.getNote() < 0 ) {
                rest(n.getLength());
                continue;
            }
            midiSoundNote(n.getNote(), n.getLength(), n.getVelocity());
        }
    }
    
    
    /**
     * Toistaa ASCII-tiedoston.
     * @param filename soitettavan tiedoston nimi
     */
    public void playAsciiFile(final String filename) {
        try ( Scanner sc = new Scanner(new File(filename)) ) {
            // Toistetaan sisältä
            String notes = sc.next();
            play(notes);
        } catch (FileNotFoundException e) {
            error("Tiedostoa " + filename + " ei läytynyt!");
        }
    }

    
    /**
     * Toistaa MIDI-tiedoston niin että ohjelma vi jatkaa muita tehtäviä.
     * Tälläin palautettu sequencer pitä ottaa vastaan ja sanoa sille
     * joskus close.
     * @param filename Tiedoston nimi, oltava midi-tiedosto
     * @return seuencer jolla soitetaan, niin voi pysäyttää milloin haluaa
     */
    public Sequencer playMidiFileAsync(final String filename) {
        try {
            // Lataa MIDI-tiedosto 
            final Sequence song = MidiSystem.getSequence(new File(filename));
            // Lataa sekvensseri
            Sequencer sequencer = MidiSystem.getSequencer();
            sequencer.open();
            sequencer.setSequence(song);
            
            if ( isVerbose() ) printSequence(song);
            
            // Aloita soitto
            sequencer.start();
            message("Soitetaan MIDI-tiedosto: " + filename);
            return sequencer;
        } catch (InvalidMidiDataException e) {
            error("Virhe luettaessa MIDI-dataa! Annoithan MIDI-tiedoston? (.mid)");
        } catch (IOException e) {
            error("Virhe avattaessa MIDI-tiedostoa: " + e.getMessage());
        } catch (MidiUnavailableException e) {
            error("MIDI-soitinta ei voida alustaa.");
        }
        return null;
    }
    
    
    /**
     * Soittaa valitun midi-tiedoston ja odottaa soiton loppumista.
     * @param filename soitettavan midi-tiedoston nimi
     */
    public void playMidiFile(final String filename) {
        try ( Sequencer sequencer = playMidiFileAsync(filename) ) {
            if ( sequencer == null ) return;
            while ( sequencer.isRunning() ) {
                try {
                    Thread.sleep(200);
                } catch (InterruptedException e) {
                    //
                }
            }
        }
    }

        
    private String messageInfotoString(MidiMessage message) {
        if (! ( message instanceof ShortMessage ) ) return message.toString();
        ShortMessage m = (ShortMessage)message;
        return String.format("Channel %d  command: %d data %d %d",m.getChannel(),m.getCommand(),m.getData1(), m.getData2());
    }


    
    /**
     * Tulostetaan nuottien tiedot
     * @param song tutkittava jono
     */
    public void printSequence(Sequence song) {
        // Print sequence information
        System.out.println( " length: "   +  song.getTickLength() + " ticks" );
        System.out.println( " duration: " +  song.getMicrosecondLength() +  " micro seconds" );
        //System.out.println( " division type: " +  divisionTypeToString( song.getDivisionType() );
        System.out.println( " division type: " +   song.getDivisionType() );
        System.out.println( " resolution type: " +   song.getResolution() );

        // Print track information
        Track[] tracks = song.getTracks();
        if (tracks == null) return;
        for (int i = 0; i < tracks.length; i++) {
            System.out.println("Track " + i + ":");
            Track track = tracks[i];
            for (int j = 0; j < track.size(); j++) {
                MidiEvent event = track.get(j);
                System.out.printf(" tick %8d, %s%n",event.getTick(),messageInfotoString(event.getMessage()));
            } // for
        } // for
    }
    
    /**
     * Toistaa neljäsosanuotin neljännessä oktaavissa.
     * @param note Nuotti (C-B)
     */
    @Override
    public void playSingleNote(final String note) {
        playPreciseNote(note, 0.25, getDefOctave());
    }
    
    /**
     * Toistaa nuotin neljännessä oktaavissa.
     * @param note Nuotti 
     */
    @Override
    public void playSingleNote(final String note, final double length) {
        playPreciseNote(note, length, getDefOctave());
    }
  
    /**
     * Toistaa minkä tahansa kestoisen nuotin tietyssä oktaavissa.
     * @param note Nuotti
     * @param length Kesto
     * @param octave Oktaavi
     */
    @Override
    public void playSingleNote(String note, double length, int octave){
        playPreciseNote(note, length, octave);
    }
    
    /**
     * Toistaa minkä tahansa kestoisen nuotin tietyssä oktaavissa.
     * @param note Nuotti
     * @param length Kesto
     * @param octave Oktaavi
     * @param velocity1 voimakkuus
     */
    @Override
    public void playSingleNote(String note, double length, int octave, int velocity1){
        playPreciseNote(note,length, octave, this.velocity);
    }
    
    
    /** 
     * Toistaa mielivaltaisen tauon.
     * @param barLength tauon pituus ms
     */
    protected void rest(final int barLength) {
        // Ei negatiivisia taukoja
        if (barLength < 0)
            return;
        try {
            int restLength = barLength;
            message("Soitin nuotin " + Note.message(Note.DELAY, barLength, Note.DELAY));
            Thread.sleep(restLength);
        } catch (InterruptedException e) {
            //
        }
    }

    
    /**
     * Valitsee kanavan. Kanavia on MIDI-spesifikaatiossa 16.
     * @param channel Kanava väliltä 1-16.
     */
    public void selectChannel(final int channel) {
        if (channel < channels.length && channel > 0) {
            currentChannel = channel;
            message("Valittiin kanava: " + channels[currentChannel]);
        } else {
            System.err.println("Virheellinen kanava, täytyy olla väliltä 1-16! Annoit: " + channel);
        }
    }
    
    /**
     * Vaihtaa instrumentin. Ks. <a href="http://www.midi.org/techspecs/gm1sound.php#instrument">spesifikaatio</a>.
     * @param instrument Instrumentti väliltä 1-128.
     */
    public void setInstrument(final int instrument) {
        checkSynth("setInstrument");
        if ( instrument < 1 || instruments.length < instrument ) {
            error(String.format("Instrumentti oltava väliltä 1-%d, annoit: %d ",instruments.length, instrument));
            return;
        } 
        message("Ladattiin soitin " + instruments[instrument - 1]);
        synth.loadInstrument(instruments[instrument - 1]);
        this.instrument = instrument;
        channels[currentChannel].programChange(instrument - 1);
    }
    
    /**
     * @return nykyinen instrumentti
     */
    public int getInstrument() {
        return instrument;
    }
    
    /**
     * Vaihtaa instrumentin. Ks. <a href="http://www.midi.org/techspecs/gm1sound.php#instrument">spesifikaatio</a>.
     * @param instrument Instrumentti väliltä 1-128.
     * @param def Oletusarvo mitä käytetään jos annettu indeksi on väärin 
     */
    public void setInstrument(final int instrument, final int def) {
        int i = instrument;
        if ( i < 1 || instruments.length < i ) i = def;
        setInstrument(i); 
    }
    
    /**
     * Asettaa tempon. BPM on iskuja minuutissa, oletus on 120. Tämä määrää syntetisaattorin tempon.
     * @param bpm asetettava tempo
     */
    @Override
    public void setTempo(final int bpm) {
        if (bpm > 0) {
            beatLength = 60000 / bpm;
            this.bpm = bpm; 
        }
    }

    
    /**
     * @return nykyinen tempo bmp
     */
    public int getTempo() {
        return bpm;
    }
    
    
    /**
     * Asetetaan nopeus jolla "kosketinta lyödään"
     * @param velocity lyöntinopeus, 64 on normaali
     */
    public void setVelocity(int velocity) {
        this.velocity = velocity;
    }

    
    /**
     * @return nopeus jolla kosketinta lyödään
     */
    public int getVelocity() {
        return velocity;
    }
    
    
    /**
     * Debug-viestit päälle/pois
     * @param set asetetaanko viestit päälle (true) vai ei (false)
     */
    public void setVerbose(final boolean set) {
        this.verbose = set;
    }
     
    /**
     * Toistaa yksittäisen nuotin instrumentilla. Standardissa MIDI 1.0-spesifikaatiossa on 86 eri nuottia,
     * jotka menevät A'' - c'''' (A0 - C8). Toisin sanoen vajaa 8 oktaavia. Nuotit ovat
     * A, Bb, B, C, C#, D, D#, E, F, F#, G, G#.
     * 
     * 
     * Nuotin kestot ovat tahdin kestoja 1/16, 1/8, 1/4, 2/4, 3/4, 4/4 ... n desimaalimuodossa eli
     * 0.0625, 0.125, 0.25, 0.5, 0.75, 1, ..., n. Tahdin pituus määräytyy tempon perusteella.
     * 
     * @param note Nuotti (ks. yllä)
     * @param octave Oktaavi (0-8).
     * @param duration Nuotin kesto iskuina.
     * @param velocity1 voimakkuus
     */
    private void playPreciseNote(final String note, final double length, final int octave, final int velocity1) {
        Note n = makeNote(note, length, octave, velocity1);
        // Kaikki ok. Soitetaan!
        midiSoundNote(n.getNote(), n.getLength(), n.getVelocity());
    }
    

    /**
     * Soittaa nuotin 
     * @param note Nuotti (ks. yllä)
     * @param length Nuotin kesto iskuina.
     * @param octave Oktaavi (0-8).
     */
    private void playPreciseNote(final String note, final double length, final int octave) {
        playPreciseNote(note, length, octave, this.velocity);
    }
    
    
    /**
     * Sulkee kanavalta nuotin.
     * @param note nuotti
     */
    protected void channelOff(final int note)
    {
        this.channels[currentChannel].noteOff(note);
    }
    
    /**
     * Avaa kanavalta nuotin.
     * @param note nuotti
     * @param velocity1 voimakkuus
     */
    protected void channelOn(final int note, final int velocity1)
    {
        this.channels[currentChannel].noteOn(note, velocity1);
    }
    
    /**
     * Tulostaa virheviestin. Tähän tarkoitukseen olisi sopinut kutsuvan metodin nimen onkiminen pinosta,
     * mutta se on hidasta. 
     * @param origin Kutsuva metodi
     */
    protected void checkSynth(final String origin) {
        if (synth == null) {
            error(origin + ": Syntetisaattori käynnistämättä!");
        }
    }
    
    /**
     * Tulostaa virheviestin ja lopettaa ohjelman.
     * @param msg Viesti
     */
    protected void error(final String msg) {
        // Selvitä kutsuneen metodin nimi
        final String callerName = Thread.currentThread().getStackTrace()[2].getMethodName();
        System.err.println("Virhe: " + callerName + ": " + msg);
        System.exit(0);
    }
    
    
    /**
     * @return iskun pituus
     */
    protected final int getBeatLength() {
        return beatLength;
    }
    
    /**
     * Tulostaa viestin mikäli viestintulostus on päällä. Oletus on, että järjestelmä on hiljaa.
     * @param msg Viesti
     */
    protected void message(final String msg) {
        if (verbose)
            System.out.println(msg);
    }
    
    /**
     * Tulostaa kaikkien instrumenttien nimet
     */
    public void printInstruments() {
        int n=1;
        for (Instrument i:instruments)
            System.out.printf("%3d %s%n",n++,i.getName());
    }
 
    
    /**
     * Etsii ensimmäisen instrumentin jonka nimi täsmää hakuehtoon
     * @param regexp regular expression hakuehto instrumentille
     * @return sen instrumentin indeksi, joka täsmää hakuehtoon, -1 = ei löydy
     */
    public int getInstrumentIndex(String regexp) {
        for (int i=0; i<instruments.length; i++) {
            Instrument inst = instruments[i];
            if ( inst.getName().trim().matches(regexp)) return i+1;
        }    
        return -1;
        
    }

    /**
     * @return onko aputulosteet päällä vai ei
     */
    public boolean isVerbose() {
        return verbose;
    }
    
}
