package fi.jyu.mit.Music;

/**
 * Luokka nuottien esittämiseen.
 */
public class Note {
    /** Taukoa kuvaava arvo */
    public static final int DELAY = -10000;

    private int note;
    private int length;
    private int velocity;
    
    /**
     * Alutetaan nuotti
     * @param note MIDI-nuotin arvo (21-108)
     * @param length Kesto millisekunteina
     * @param velocity Nuotin nk. painovoimakkuus. Oletus on 64.     
     */
    public Note(int note, int length, int velocity) {
        this.setNote(note);
        this.setLength(length);
        this.setVelocity(velocity);
    }

    
    /**
     * Taukoa kuvaava nuotti
     * @author vesal
     * @version 13.11.2010
     */
    public static class Delay extends Note {
        /**
         * Muodotetaan uusi tauko
         * @param length tauon pituus
         * @example
         * <pre name="test">
         *    Note n = new Note.Delay(200);
         *    n.isDelay() === true;
         *    n.getLength() === 200;
         * </pre>
         */
        public Delay(int length) {
            super(DELAY,length,DELAY);
            super.setNote(DELAY);
            super.setVelocity(DELAY);
        }

        
        /**
         * Ei tee mitään tauko-luokassa
         * @example
         * <pre name="test">
         *    Note n = new Note.Delay(200);
         *    n.setNote(60);
         *    n.isDelay() === true;    
         * </pre>
         */
        @Override
        public void setNote(int note) {
            // tauossa ei ole nuottia
        }
        
    }
    
    
    /**
     * Tehdään nuotista kopio
     * @param n nuotti jolla uusi nuotti alustetaan.
     */
    public Note(Note n) {
        this(n.note,n.length,n.velocity);
    }
    
    
    /**
     * Alustetaan nuotin arvo. Negatiivinen on tauko.
     * @param note MIDI-nuotin arvo (21-108)
     * @example
     * <pre name="test">
     *   Note n = new Note(30,300,64);
     *   n.getNote() === 30;
     *   n.setNote(Note.DELAY);
     *   n.isDelay() === true;
     *   n.setNote(2);
     *   n.getNote() === 21;
     *   n.setNote(200);
     *   n.getNote() === 108; 
     *   n.setNote(-1);
     *   n.isDelay() === true;
     * </pre>
     */
    public void setNote(int note) {
        this.note = note;
        if ( note < 21 ) this.note = 21;
        if ( note > 108 ) this.note = 108;
        if ( note < 0 ) { this.note = DELAY; this.velocity = DELAY; }
    }

    
    /**
     * @return MIDI-nuotin arvo (21-108)
     */
    public int getNote() {
        return note;
    }
    
    
    /**
     * @return true jos nuotti edustaa taukoa
     */
    public boolean isDelay() {
        return getNote() < 0;
    }
    
    
    /**
     * asetetaan nuotin pituus
     * @param length asetettava pituus
     */
    public void setLength(int length) {
        this.length = length;
    }
    
    
    /**
     * @return nuotin pituus
     */
    public int getLength() {
        return length;
    }
    
    
    /**
     * Asetetaan kuinka voimakkaasti painetaan. 
     * @param velocity Nuotin nk. painovoimakkuus. Oletus on 64.     
     */
    public void setVelocity(int velocity) {
        this.velocity = velocity;
    }
    
    
    /**
     * @return nuotin nk. painovoimakkuus. Oletus on 64.
     */
    public int getVelocity() {
        return velocity;
    }
    
    
    /**
     * Muuttaa modi-nuotin merkkijonoksi.  Tauko on "-" 
     * @param note MIDI-nuotin arvo (21-108)
     * @return midi-nuotin arvo merkkijonona.
     * 
     * @example
     * <pre name="test">
     *   Note.toString(21) === "A0";
     *   Note.toString(22) === "Bb0";
     *   Note.toString(23) === "B0";
     *   Note.toString(24) === "C1";
     *   Note.toString(25) === "C#1";
     *   Note.toString(60) === "C4";
     *   Note.toString(106) === "Bb7";
     *   Note.toString(107) === "B7";
     *   Note.toString(108) === "C8";
     *   Note.toString(-1) === "-";
     * </pre>
     */
    public static String toString(int note) {
        if (note < 0) 
            return "-";
        StringBuffer sb = new StringBuffer();
        int base = 24;
        int noteIndex = 0;
        int octave = 0;
        // Jos nuotti > 23, ollaan nollaoktaavin puolella, jolloin pitää oktaavien lukumäärä vähentää nuotista.
        if (note > 23) { 
            octave = (note / 12) - 1;
            noteIndex = note - (base + ((octave - 1) * 12));
        } else {
            noteIndex = note - 12;
        }
        if (noteIndex != 10) {
            sb.append(MidiPlayer.ALLNOTES[noteIndex]);
        } else {
            sb.append("Bb");
        }
        // Lisää oktaavi
        sb.append(octave);
        return sb.toString();
    }
    
    
    /**
     * Muuttaa nuotin merkkijonoksi
     * @return nuotti merkkijonona
     * @example
     * <pre name="test">
     *   Note n = new Note(60,250,64);
     *   n.toString() === "C4";
     * </pre>
     */
    @Override
    public String toString() {
        return toString(note);
    }

    
    /**
     * Palauttaa nuottia vastaavan tulostettavan viestin
     * 
     * @param note MIDI-nuotin arvo (21-108)
     * @param length Kesto millisekunteina
     * @param velocity Nuotin nk. painovoimakkuus. Oletus on 64.     
     * @return nuottia vastaava "debug"-viesti
     * 
     * @example
     * <pre name="test">
     *   Note.message(60,250,64) === "' 60' C4  kesto  250 ms";  
     *   Note.message(-1,250,64) === "'   '     kesto  250 ms";  
     * </pre>
     */
    @SuppressWarnings("unused")
    public static String message(int note, int length, int velocity) {
        if ( note < 0 )
            return String.format("'%3s' %-3s kesto %4d ms", "","",length);
          return String.format("'%3d' %-3s kesto %4d ms", note,toString(note),length);
        
    }
    
    /**
     * Palauttaa nuottia vastaavan tulostettavan viestin
     * @return nuotti jonona "' 60' C4  kesto  250 ms"  
     * 
     * @example
     * <pre name="test">
     *   Note n = new Note(60,250,64);
     *   n.message() === "' 60' C4  kesto  250 ms";
     * </pre>
     */
    public String message() {
        return message(getNote(),getLength(),getVelocity());
    }
    
}



