package fi.jyu.mit.Music;

import java.util.LinkedList;
import java.util.List;

import java.util.Vector;

/**
 * Tarjoaa nuottien rinnakkaisen toistomahdollisuuden, jossa toistot tapahtuvat eri säikeissä, mikä mahdollistaa
 * ns. ääniraitojen teon kutsumalla montaa eri play-komentoa samanaikaisesti.
 */
public class ThreadedMidiPlayer extends MidiPlayer {
    
    private class SoundConsumer extends Thread {
        // @SuppressWarnings("synthetic-access")
        @Override
        public void run() {
            Clock c = new Clock();
            while (true) {
                Note n = null;
                synchronized(queue) {
                    // odota pysäytettäviä nuotteja
                    while (queue.size() == 0) {
                        if (stop)
                            break;
                        try {
                            queue.wait(); // Odotetaan lukkoa
                        } catch (InterruptedException ie) {
                          //  
                        }
                    }
                    if (stop) 
                        break;
                    n = queue.remove(0);
                    queue.notifyAll(); 
                    
                }
                // Pysäytä nuotti. ei ole lukollinen
                message(String.format("%6d  Pysäytin   %s",c.getElapsed(),n.message()));
                channelOff(n.getNote());
            }
            
        }
    }
    /**
     * Laittaa nuotit pysäytysjonoon.
     */
    private class SoundProducer extends Thread {
        private double noteLength;
        private Vector<Note> notes;

        public SoundProducer() {
            noteLength = 0.25;
            notes = new Vector<Note>();
        }

        public void addNote(Note note) {
            notes.add(note);
        }
        
        // Käynnistää nuotintuottaja-säikeen
        // @SuppressWarnings("synthetic-access")
        @Override
        public void run() {
            Clock c = new Clock();
            long breakLength = (long)((getBeatLength() * 4) * noteLength);
            if (notes.size() == 0) 
                return;
            Vector<Note> played = new Vector<Note>();
            while (true) {
                if (notes.isEmpty())
                    break;
                played.clear();
                while (!notes.isEmpty()) {
                    Note n = notes.firstElement();
                    if ( !n.isDelay() ) {
                        message(String.format("%6d  Käynnistän %s",c.getElapsed(),n.message()));
                        played.add(n);
                        channelOn(n.getNote(), n.getVelocity());
                    } else {
                        notes.remove(0);
                        break;
                    }
                    notes.remove(0);
                }
                
                // Tauko. Jos soitettiin yksi (tai useampi) nuotti niin odotetaan, tai jos tuli tauko.
                try {
                    Thread.sleep(breakLength);
                } catch (InterruptedException ie) {
                    //
                }
                    
                // Tämä edellyttää lukkoa
                for (Note n : played) {
                    synchronized(queue) {         
                        // Soiton jälkeen lisätään nuotti pysäytettävien nuottien jonoon
                        // message(c.getElapsed() + " Laitan pysäytysjonoon " + n.note + ", kesto " + n.length + " ms");
                        queue.add(new Note(n.getNote(), n.getLength(), n.getVelocity()));
                        queue.notifyAll(); // nyt on nuotteja toistettavana
                    }
                }
            }
        }
        
        protected void setNoteLength(double noteLength) {
            this.noteLength = noteLength;
        }
    }
    class Clock {
        long startMillis = 0;
        public Clock() {
            startMillis = System.currentTimeMillis();
        }
        public long getElapsed() {
            return (System.currentTimeMillis() - startMillis);
        }
    }
    private Vector<SoundProducer> producers;
    
    private LinkedList<Note> queue = new LinkedList<Note>();
    
    private SoundConsumer sc;
    
    private boolean stop = false; // lopetuslippu
    
    /**
     * Alustetaan säikeistetty soitin
     */
    // @SuppressWarnings("synthetic-access")
    public ThreadedMidiPlayer() {
        init();
        sc = new SoundConsumer();    
        producers = new Vector<SoundProducer>();
    }
    
    /**
     * Käynnistää soittajan. Tämän kutsuminen on välttämätäntä, jotta nuotit soisivat.
     */
    public void begin() {
        sc.start();
    }
    
    /**
     * Lopettaa soiton ja pysäyttää kaikki säikeet.
     */
    public void end() {
        synchronized(queue) {
            stop = true;
            queue.notifyAll();
        }
    }
    
    /**
     * Toistaa saamansa nuotit samanaikaisesti. Parametrilla "AAA AAA" soitetaan ensiksi kolme samanaikaista A-nuottia
     * jonka jälkeen on pituuden määräämä tauko, jonka jälkeen soitetaan loput AAA. Soitto aloitetaan heti.
     */
    @Override
    public void play(String sequence, double length, int octave, int velocity) {
       List<Note> notes = getNotes(sequence, length, octave, velocity);
        SoundProducer sp = getFreeProducer();
        for (Note n : notes)
            sp.addNote(n);
        sp.setNoteLength(length);
        sp.start();
    }
    
    
    /**
     * Toistaa kaiken samanaikaisesti kuten play-metodissa mutta tauot jätetään huomioimatta.
     * @param chord Sointu, esim. "C5E5G5"
     * @param length Soinnun pituus tahdin desimaaliosana (0.25 = neljäsosanuotti)
     * @param octave Soinnun oletusoktaavi. Jos chord-parametrissä on nuotteja, joissa ei ole oktaavia määriteltynä, käyttää soitin tätä oktaavia.
     * @param velocity voimakkuus
     */
    public void playChord(String chord, double length, int octave, int velocity) {
        List<Note> notes = getNotes(chord, length, octave, velocity);
        SoundProducer sp = getFreeProducer();
        for (Note n : notes) {
            // Hypätään taukojen yli    
            if ( !n.isDelay() )                     
                sp.addNote(n);
        }
        sp.setNoteLength(length);
        sp.start();
        try {
            Thread.sleep((long)((getBeatLength()*4) * length));
        } catch (InterruptedException e) {
            //
        }
    }
    
    /**
     *  tauko, joka ei tee tässä yhteydesä mitään
     */
    public void rest() { /* */ }
    
    /**
     * @return tuottaja
     */
    protected SoundProducer getFreeProducer() {
        // Poistetaan toimimattomat tuottajat
        for (int i = 0; i < producers.size(); i++) {
            if (!producers.elementAt(i).isAlive()) {
                producers.remove(i);
            }
        }
        // Luodaan uusi
        SoundProducer sp = new SoundProducer();
        producers.addElement(sp);
        return sp;
    }
}
